# Hobbits Plugin Development

The Hobbits framework is designed to coordinate the capabilities of various
plugins into a smooth, cohesive experience. The plugins are where all of the
"actual work" gets done.

There are currently four types of plugins in Hobbits: Displays, Analyzers,
Operators, and Importer/Exporters. This document will cover general plugin
implementation concepts, and then walk through the development process for each
type of plugin.

## Example Plugin Development Video

![YOUTUBE](Dg3vknwLO74)

## General Terms and Concepts

### Qt Plugin API

Hobbits leverages the low-level plugin API provided by Qt for its plugin system.
Refer to the following Qt documentation for details:

 - [Low-level Plugin API](https://doc.qt.io/qt-5/plugins-howto.html#the-low-level-api-extending-qt-applications)
 - [Deploying Plugins](https://doc.qt.io/qt-5/deployment-plugins.html)

### Bit Containers

The `BitContainer` class is the primary means of storing data, and sharing it
between plugins. A `BitContainer` holds binary data, as well as metadata such as
how that data is framed. It also keeps a record of any changes that have been
made to it by Hobbits plugins.

The binary data is held in a `BitArray`, which abstracts byte boundaries from
the user, and has a modest-but-growing assortment of performance efficiencies
and helper functions.

Metadata is stored in a `BitInfo` object which is read and written with the
`bitInfo` and `setBitInfo` methods.

The framing of a bitstream is an important part of its metadata for display
and processing purposes. The frames of a `BitContainer` can be set in the
`BitInfo::setFrames` method.

General purpose metadata can be set with `BitInfo::setMetadata` and read with
`BitInfo::metadata`. Plugins
can use container metadata as a general purpose bus for communicating with other
plugins, or even as a cache for their own purposes. For example, the "Find"
Analyzer plugin stores the results of a find operation in a container's
metadata so that when that container loses focus, the information will still be
available when the container comes back into focus.

Highlights are a separate kind of metadata that deal with ranges of bits within
a bit container. They are read and written with various functions including
`BitInfo::addHighlights` and `BitInfo::highlights`
, and they are stored as a map of strings to `RangeHighlight` lists.
The "Find" Analyzer plugin
sets highlights in the ranges where it finds its search string, and several
Display plugins check for the existence and colors of `RangeHighlight`s in order
to render visual highlights across those ranges. Plugins like the `Extractor`
operator use `RangeHighlight`s for their processing.

Analyzer plugins do not have write access to bit containers, so they cannot edit
the `BitInfo` directly. As a result, they must include
a container's updated `BitInfo` in the
`AnalyzerResult`s that they return.

### Common Plugin Interface Methods

All of the plugin interfaces have a `getName` method that must return the name
of that plugin as it should appear to the user. The name must be unique. There
is also a common `createDefaultX` method (where X is the interface type) that
simply returns a default instance of the plugin.

The Analyzer and Operator interfaces have several simple methods in common:

 - `applyToWidget` initializes the plugin's UI on a given widget
 - `provideCallback` gives the plugin a `PluginCallback` instance that can be
   used to request that it gets executed (this is useful because the central
   application that triggers plugin execution is decoupled from the plugin's UI,
   so in order for a plugin to trigger execution when, for example, the "Enter"
   key is pressed in its UI form, it needs to relay this request through the
   `PluginCallback` instance it is given.)
 - `getStateFromUi` gets the plugin's current state/configuration based on how
   its UI is filled out. If there are invalid or incomplete entries, the plugin
   can return an empty state to indicate that it is not ready for execution.
 - `setPluginStateInUi` (conversely from `getStateFromUI`) is given a
   pluginState, and it must fill out the UI in a way that would return that
   state if `getStateFromUi` were called immediately afterwards
 - `canRecallPluginState` checks a plugin state to see if it can be used to
   execute the plugin. If the state is empty or incomplete, it must return
   false.
 - `previewBits` takes a read-only bit container and it can be used to prepare the
   plugin for execution or enhance the UI options it presents. The Width Framer
   analyzer plugins generates auto-correlation data in the `previewBits` method,
   which is used to display a width-selector graph with suggested widths. It is
   fine to leave this method's implementation empty.

#### Plugin State

The plugin state contains parameters for how the plugin should run. For example,
the Take Skip operator plugin will require a "take_skip_string" parameter to
guide its execution. The state that it passed in can be provided by the plugin's
own `getStateFromUi` method, or it can come from a saved batch that was
generated from a state returned by a plugin result (`OperatorResult` or
`AnalyzerResult`.) Plugin results will usually just return the state that was
passed in (e.g. `AnalyzerResult::result(bitInfo, recallablePluginState)`,) but
sometimes
that state needs to be enriched with execution details that were added by the
plugin, such as a random number seed, so that the execution can be duplicated
exactly with the returned state.

#### Action Progress

The `ActionProgress` parameter is passed to the primary methods of the operator
and analyzer interfaces. It allows a plugin to report its percentage progress as
it executes, and allows it to check if it has been cancelled so that it can halt
execution gracefully.

## Specific Interfaces

### Operator Interface

Operators take some number of data inputs, and produce some number of data
outputs. Most operators are 1-in, 1-out, but the flexibility of the interface
enables several critical operations (e.g. data generators, muxes, demuxes.)
The output of operators will go into new containers so that the original data
can still be easily referenced.

`operateOnContainers` takes a list of read-only bit containers, a plugin state,
and an `ActionProgress` instance. It returns an `OperatorResult` which contains
any new bit containers that have been created by the operator, and the plugin
state that will enable the operation to be duplicated exactly (see 
[plugin state details above](#plugin-state).) **`operateOnContainers` might run
in a separate thread from the main GUI thread, so you cannot use Qt features
that depend on being in the main thread, e.g., showing a QMessageBox.**

`getMinInputContainers` and `getMaxInputContainers` return the minimum and
maximum number of containers that the operator accepts as inputs. Operators that
take a single input and produce a single output would simply return 1 for both
of these functions.

### Analyzer Interface

Analyzers digest and decorate the data in a way that facilitates follow-on
processing and/or human evaluation.

`analyzeBits` takes a read-only bit container, a plugin state, and an
`ActionProgress` instance. It returns an `AnalyzerResult` which contains new
general metadata and range entries for the container, and the plugin
state that will enable the operation to be duplicated exactly (see 
[plugin state details above](#plugin-state).) **`analyzeBits` might run in a
separate thread from the main GUI thread, so you cannot use Qt features that
depend on being in the main thread, e.g., showing a QMessageBox.**

### Display Interface

Displays show the data in some sort of useful format (e.g. a bit raster, or a
hex dump.)  They can use metadata and highlights in bit containers to augment a
depiction of the data, but they could also show *only* the metadata if that was
useful somehow (e.g. if there were some metrics worth presenting separately from
the plugin that generated them.)

`getDisplayWidget` and `getControlsWidget` provide the display and (optional)
controls for a display plugin. They both take a `DisplayHandle` as their sole
parameter.

The `DisplayHandle` provides access to a variety of things that a display might
need to access (e.g. the current bit container), or set (e.g. the bit that is
currently being hovered over by the mouse.)

### Import/Export Interface

Import/Export plugins import and/or export bit containers for use in Hobbits.
The simplest of these plugins is the "File Data" plugin that imports/exports
raw binary data from/to files.

`canExport` and `canImport` return a boolean value indicating whether or not the
plugin is capable of importing or exporting (e.g. you might want a plugin that
can import data but not export it.) You can return `false` in both methods, but
that would be impolite.

`importBits` and `exportBits` both receive a map of arguments (which might be
empty) and a pointer to a `QWidget` that should be used as the parent to any GUI
elements generated by the plugin (e.g. a file dialog.) 

`exportBits` also receives a bit container that should be exported.

`importBits` returns a non-empty bit container if it was successful at importing
data.

## Helpful Tools

In order to simplify the process of developing plugins, there are Qt Creator
wizards that take care of most of the boilerplate code.

The wizards can be installed using the wizard_installer.sh script.

## Get in touch

If you are having trouble making a plugin, or if you have a suggestion for
the hobbits API or documentation please open an issue on
[the GitHub page](https://github.com/Mahlet-Inc/hobbits/issues)
or ask about it on [the Discord channel](https://discord.gg/wRQJpZZ)
